/*
 * Copyright (c) 2013 CNRS/LAAS
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * Documentation for individual commands is extracted from the 
 * RMP 440 LE User Manual 
 */

#include <stdint.h>
#include <stdio.h>

#include <rmp440/rmp440.h>

#include "io.h"
/*
 * Basic motion
 *
 * parameters: v: velocity, w: yaw range -1.0..1.0
 *
 * the motion command packet is used to command the machine velocity
 * and yaw rate. The commands are normalized (-1.0..1.0). The
 * normalized valudes are scaled against the user configurable
 * parameters associated with the controller. The parameter against
 * which the command is scaled depends on the input mapping type. See
 * the documentation on Controller Input Mapping p. 54 for details.
 */
int
rmp440Motion(const struct rmp440_io *rmp, double vel, double yaw)
{
	return rmp440SendMsgFloat(rmp, RMP_CAN_CMD_ID, vel, yaw);
}

/*
 * This command can be used to poll the CCU for data without issuing a
 * command that will result in an action. This command does nothing,
 * but it is valid and will sollicit a response.
 */
int
rmp440CmdNone(const struct rmp440_io *rmp)
{
	return rmp440SendMsgInt(rmp, RMP_CMD_NONE, 0, 0);
}

/*
 * rmp440SetMaximumVelocity 
 * Parameter vel: max velocity. Unit m/s
 *
 * This command is used to set the user defined maximum velocity
 * limit.
 * See section XX on velocity controller input mapping for how this
 * value will affect velocity commands.
 */
int
rmp440SetMaximumVelocity(const struct rmp440_io *rmp, double vel)
{
	if (vel < RMP_MIN_VELOCITY || vel > RMP_MAX_VELOCITY)
		return -1;
	return rmp440SendMsgIntFloat(rmp, RMP_CFG_CMD_ID,
	    RMP_CMD_SET_MAXIMUM_VELOCITY, vel);
}

/*
 * rmp440SetMaximumAcceleration
 * Parameter acc: max acceleration. Unit m/s²
 *
 * This command is used to set the user defined maximum acceleration
 * limit.
 * See Section on velocity controller input mapping for how this value
 * will affect velocity commands 
 */
int
rmp440SetMaximumAcceleration(const struct rmp440_io *rmp, double acc)
{
	if (acc < RMP_MIN_ACCEL || acc > RMP_MAX_ACCEL)
		return -1;
	return rmp440SendMsgIntFloat(rmp, RMP_CFG_CMD_ID,
	    RMP_CMD_SET_MAXIMUM_ACCEL, acc);
}

/*
 * rmp440SetMaximumDeceleration
 * parameter: dec: max decelleration. Unit m/s²
 *
 * This command is used to set the user defined maximum decceleration
 * limit.
 * See the section on velocity controller input mapping for how this
 * value will affect velocity commands
 *
 * WARNING: setting this value to zero will result in the machine not
 * being able to stop.
 */
int
rmp440SetMaximumDeceleration(const struct rmp440_io *rmp, double dec)
{
	if (dec < RMP_MIN_DECEL || dec > RMP_MAX_DECEL)
		return -1;
	return rmp440SendMsgIntFloat(rmp, RMP_CFG_CMD_ID,
	    RMP_CMD_SET_MAXIMUM_DECEL, dec);
}

/*
 * rmp440SetMaximumDTZDecelRate
 * parameter: dec: max deceleration rate. Units m/s²
 *
 * This command is used to set the user defined maximum decel to zero
 * response deceleration rate.
 * When a decel to zero respons is command via mode command, through
 * hardware, or through fault response, this is the maximum rate at
 * which the machine will come to a stop.
 *
 * WARNING: setting the maximum DTZ deceleration to zero will result
 * in the machine not being able to stop in a decel response.
 */
int
rmp440SetMaximumDTZDecelRate(const struct rmp440_io *rmp, double dec)
{
	if (dec < RMP_MIN_DTZ_RATE || dec > RMP_MAX_DTZ_RATE)
		return -1;
	return rmp440SendMsgIntFloat(rmp, RMP_CFG_CMD_ID,
	    RMP_CMD_SET_MAXIMUM_DTZ_RATE, dec);
}

/*
 * rmp440SetCoastDownAccel
 * parameter: acc: coastdown acceleration. Unit m/s²
 *
 * This command is used to set the user defined coastdown acceleration
 * value for acceleration based input mapping for the velocity
 * controller. See the section on velocity controller input mapping
 * for how this value will affect velocity commands.
 *
 * WARNING: Setting the coastdown acceleration to zero will result int
 * the machine maintainin constant velocity with zero command when
 * using Acceleration based input mapping.
 */
int
rmp440SetCoastDownAccel(const struct rmp440_io *rmp, double acc)
{
	if (acc < RMP_MIN_COASTDOWN_ACCEL || acc > RMP_MAX_COASTDOWN_ACCEL)
		return -1;
	return rmp440SendMsgIntFloat(rmp, RMP_CFG_CMD_ID,
	    RMP_CMD_SET_COASTDOWN_ACCEL, acc);
}

/* 
 * rmp440SetMaximumTurnRate
 * parameter: rate: yaw rate limit. Unit rad/s
 *
 * This command is used to set the user defined yaw rate limit. See
 * the section on yaw controller input mapping for how this value will
 * affect yaw rate commands.
 *
 * WARNING: setting the maximum turn rate to zero will result in the
 * machine not being able to turn.
 */
int
rmp440SetMaximumTurnRate(const struct rmp440_io *rmp, double rate)
{
	if (rate < RMP_MIN_YAW_RATE || rate > RMP_MAX_YAW_RATE)
		return -1;
	return rmp440SendMsgIntFloat(rmp, RMP_CFG_CMD_ID,
	    RMP_CMD_SET_MAXIMUM_TURN_RATE, rate);
}

/*
 * rmp440SetMaximumTurnAccel
 * parameter: rate: yaw acceleration limit. Unit rad/s²
 *
 * This commnd is used to set the user defined yaw acceleration
 * limit. This value limit the rate at which the raw rate target can
 * change in the yaw controller.
 *
 * WARNING: setting the maximum turn acceleration to zero will result
 * in the machine not being able to turn
 */
int
rmp440SetMaximumTurnAccel(const struct rmp440_io *rmp, double rate)
{
	if (rate < RMP_MIN_YAW_ACCEL || rate > RMP_MAX_YAW_ACCEL)
		return -1;
	return rmp440SendMsgIntFloat(rmp, RMP_CFG_CMD_ID,
	    RMP_CMD_SET_MAXIMUM_TURN_ACCEL, rate);
}

/*
 * rmp440SetTireDiameter
 * parameter: diam: tire diameter. Unit m
 *
 * This command updates the machine tire diameter used in software to
 * calculate velocity, position, differential wheel speed (yaw rate)
 * and accelerations.
 *
 * CAUTION: this value must match the actual tire diameter on the
 * machine. Failure to do so will result in undetermined behaviour and
 * invalid feedback
 */
int
rmp440SetTireDiameter(const struct rmp440_io *rmp, double diam)
{
	if (diam < RMP_MIN_TIRE_DIAMETER || diam > RMP_MAX_TIRE_DIAMETER)
		return -1;
	return rmp440SendMsgIntFloat(rmp, RMP_CFG_CMD_ID,
	    RMP_CMD_SET_TIRE_DIAMETER, diam);
}

/*
 * rmp440SetWheelBaseLength
 * parameter: len: wheel base length. Unit m
 *
 * This command updates the machine wheel base length (longitudinal
 * distance between the tires) used in software to calculate the
 * mecanum transform for omnidirectional platforms. It is not used for
 * other platforms.
 *
 * CAUTION: this value must match the actual wheel base length on the
 * machine. Failure to do so will result in undetermined behaviour and
 * invalid feedback
 */
int
rmp440SetWheelBaseLength(const struct rmp440_io *rmp, double len)
{
	if (len < RMP_MIN_WHEEL_BASE_LENGTH || len > RMP_MAX_WHEEL_BASE_LENGTH)
		return -1;
	return rmp440SendMsgIntFloat(rmp, RMP_CFG_CMD_ID,
	    RMP_CMD_SET_WHEEL_BASE_LENGTH, len);
}

/*
 * rmp440SetWheelTrackWidth
 * parameter: width: machine track width. Unit m
 *
 * This command updates the machine track width (lateral distance
 * between the tires) used in software to calculate differential wheel
 * speeds (yaw rate) and lateral accelerations.
 *
 * CAUTION: this value must match the actual track width on the
 * machine. Failure to do so will result in undetermined behaviour and
 * invalid feedback
 */
int
rmp440SetWheelTrackWidth(const struct rmp440_io *rmp, double width)
{
	if (width < RMP_MIN_WHEEL_TRACK_WIDTH ||
	    width > RMP_MAX_WHEEL_TRACK_WIDTH)
		return -1;
	return rmp440SendMsgIntFloat(rmp, RMP_CFG_CMD_ID,
	    RMP_CMD_SET_WHEEL_TRACK_WIDTH, width);
}

/*
 * rmp440SetTransmissionRatio
 * Parameter ratio: gearbox ratio. Unitless
 *
 * This command updates the machine gearbox (transmission) ratio. It
 * is used in software to convert from motor speed to gearbox output
 * speed. 
 *
 * CAUTION: this value must match the actual gear ratio on the
 * machine. Failure to do so will result in undetermined behaviour and
 * invalid feedback
 */
int
rmp440SetTransmissionRatio(const struct rmp440_io *rmp, double ratio)
{
	if (ratio < RMP_MIN_TRANSMISSION_RATIO ||
	    ratio > RMP_MAX_TRANSMISSION_RATIO)
		return -1;
	return rmp440SendMsgIntFloat(rmp, RMP_CFG_CMD_ID,
	    RMP_CMD_SET_TRANSMISSION_RATIO, ratio);
}

/*
 * rmp440SetInputConfigBitmap
 * Parameter: bitmap
 *
 * This command updates machine behaviour configurations. It updates
 * the input mappin, audio silence setting and whether to check and
 * warn for charger present at startup.
 * When the audio silence bit is set the machine will become silent
 * and not issue audio indications.
 * For an explanation of input mapping see the Controller Input
 * Mapping section.
 */
int
rmp440SetInputConfigBitmap(const struct rmp440_io *rmp, uint32_t bitmap)
{
	return rmp440SendMsgInt(rmp, RMP_CFG_CMD_ID,
	    RMP_CMD_SET_INPUT_CONFIG_BITMAP, bitmap);
}

/*
 * rmp440SetUserFeedBack
 * parameter: bitmap[4] array of 4 feedback bitmaps
 *
 * This command updates the user defined feedback bitmaps 1 to 4. They
 * are used to select feedback from the list of variables defined in
 * appendix A do D user defined feedback bitmap 1 to 4.
 */
int
rmp440SetUserFeedBack(const struct rmp440_io *rmp, uint32_t bitmap[4])
{
	if (rmp440SendMsgInt(rmp, RMP_CFG_CMD_ID,
		RMP_CMD_SET_USER_FB_1_BITMAP, bitmap[0]) == -1)
		return -1;
	if (rmp440SendMsgInt(rmp, RMP_CFG_CMD_ID,
		RMP_CMD_SET_USER_FB_2_BITMAP, bitmap[1]) == -1)
		return -1;
	if (rmp440SendMsgInt(rmp, RMP_CFG_CMD_ID,
		RMP_CMD_SET_USER_FB_3_BITMAP, bitmap[2]) == -1)
		return -1;
	if (rmp440SendMsgInt(rmp, RMP_CFG_CMD_ID,
		RMP_CMD_SET_USER_FB_4_BITMAP, bitmap[3]) == -1)
		return -1;
	return 0;
}

/*
 * rmp440SetAudioCommand
 * parameter: song
 *
 * This command requests an audio song from the machine's motor
 * audio. If the machine deems that it is able to play the song it
 * will do so. If it is internally using the audio or the current
 * limit is folded back the machine will not plat the commanded
 * audio.
 * Audio son request should be momentary (ie only need to be sent
 * once). 
 * The song that are not persistent will be cleared by the CCU. If the
 * sont is persistent it must be cleared by sending the
 * MOTOR_AUDIO_PLAY_NO_SONG parameter.
 */
int
rmp440SetAudioCommand(const struct rmp440_io *rmp, uint32_t song)
{
	return rmp440SendMsgInt(rmp, RMP_CFG_CMD_ID,
	    RMP_CMD_SET_AUDIO_COMMAND, song);
}

/*
 * rmp440SetInetParams
 * parameters: address, port, netmask, default gateway
 *
 * This command updates machine IPv4 address, port on which it listens
 * to, netmask and default gateway.
 *
 * IP addresses and netmask are stored in uint32_t values, in host 
 * byte order (XXX to be verifed...)
 */
int
rmp440SetInetParams(const struct rmp440_io *rmp, uint32_t addr, uint16_t port,
    uint32_t mask, uint32_t gateway)
{
	if (rmp440SendMsgInt(rmp, RMP_CFG_CMD_ID,
		RMP_CMD_SET_ETH_IP_ADDRESS, addr) == -1)
		return -1;
	if (rmp440SendMsgInt(rmp, RMP_CFG_CMD_ID,
		RMP_CMD_SET_ETH_PORT_NUMBER, port) == -1)
		return -1;
	if (rmp440SendMsgInt(rmp, RMP_CFG_CMD_ID,
		RMP_CMD_SET_ETH_SUBNET_MASK, mask) == -1)
		return -1;
	if (rmp440SendMsgInt(rmp, RMP_CFG_CMD_ID,
		RMP_CMD_SET_ETH_GATEWAY, gateway) == -1)
		return -1;
	return 0;
}

/*
 * rmp440SetOperationalMode
 * parameter: operational mode
 *
 * This command is used to request mode transitions for the RMP. Each
 * mode that can be requested is listed in rmp440.h. Te persistence of
 * the request is managed internally to the CCU.
 */
int
rmp440SetOperationalMode(const struct rmp440_io *rmp, uint32_t mode)
{
	if (mode < RMP_DISABLE_REQUEST || mode > RMP_TRACTOR_REQUEST)
		return -1;
	return rmp440SendMsgInt(rmp, RMP_CFG_CMD_ID,
	    RMP_CMD_SET_OPERATIONAL_MODE, mode);
}
